/***************************************************************************
 *   Copyright (C) 2005 by Michal Vaner                                    *
 *   vorner@centrum.cz                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <vector>

#include "setupmanager.h"
using std::string;
using std::vector;

namespace vorlib {

setupmanager_e::setupmanager_e(int id, const string &message) : exception(id, message) {};

setupmanager::setupmanager() {
  automanag = false;
  file = "";//we are empty and not set to any file
}

setupmanager::setupmanager(const string &file, bool automanag) {
  this -> automanag = false;//there's no file yet, so no automanag yet
  setfile(file, automanag);//just set to proper file and let this func to load it, if automanag
}

setupmanager::~setupmanager() {
  if (automanag) save();//saving now
}

void setupmanager::setfile(const string &_file, bool _automanag) {
  if (automanag) save();//save actual file if wanted
  automanag = _automanag;
  file = _file;//set the file properly
  if (automanag) load();//and load it if needed
}

void setupmanager::load() {
  clear();
  loadfile(file);
}

void setupmanager::loadfile(const string &name) {
  FILE *f = fopen(name.c_str(), "rt");//try to open it
  if (!f) throw new setupmanager_e(1, "file not found: " + name);//didn't work
  char line[10240];//this should be enough
  char *l;//the anwser of fgets
  section *actsec = NULL;//no section yet
  while ((l = fgets(line, 10240, f)) != NULL) {//read all lines, one by one
    if (strlen(l) < 3) continue;//empty (or almost) line
    if (*l == '#') continue;//it's comment
    if (l[strlen(l) - 1] == '\n') l[strlen(l) - 1] = 0;//cut the newline at the end
    if (*l == '[') {//it's section name
      for (char *c = l + 1; *c; c++) if (*c == ']') {//search the string for ]
        *c = 0;//cut the ]
        string secname = l + 1;//get the name, skip the [ at the begining
        section newsec;//create new section
        if (!sections.add(secname, newsec))//put it in sections
          throw new setupmanager_e(3, "duplicit section name: " + secname);//duplicit name
        actsec = sections.try_get(secname);//and get pointer to it
        goto NEXTLINE;//continue with next line, it is not a value
      }
      throw new setupmanager_e(4, "syntax error, ']' missing in section name");//no ']' in the string
    }
    //it should be a value
    for (char *c = l; *c; c++) if (*c == '=') {
      if (!actsec) throw new setupmanager_e(6, "value outside a section");
      *c = 0;//split the string into 2
      string valuename = l;//the begining of it is the name
      string value = c + 1;//after the '=' begins the value
      value_t newvalue = processvalue(value);//detect with all
      actsec -> add(valuename, newvalue);//and place it there
      goto NEXTLINE;
    }
    throw new setupmanager_e(5, "syntax error, no '=' in value string");//'=' not met in the string
    NEXTLINE:;
  }
  if (fclose(f)) throw setupmanager_e(2, "file not closed: " + name);//didn't closed the file
}

void setupmanager::connectfiles(const string &_file) {
  if (automanag) save();//save it before continuing
  file = _file;//get new filename
  automanag = true;//it wouldn't work other way
  loadfile(file);//and load the new file in adition to the old one
  save();//and save it all with new file name
}

void setupmanager::save() const {
  if (file == "") throw new setupmanager_e(7, "empty file name");
  FILE *f = fopen(file.c_str(), "wt");//rewrite the file with setup
  if (!f) throw new setupmanager_e(8, "can not write to file: " + file);//it is not opened

  vector<string> sectionnames = sections.get_indexes();//get the names of sections

  for (vector<string>::iterator secname = sectionnames.begin(); secname != sectionnames.end(); secname++) {//run trough all sections
    fprintf(f, "[%s]\n", secname -> c_str());//write the section name
    const section &actsec = sections[*secname];//get the right section
    vector<string> valuenames = actsec.get_indexes();//get names of all values
    vector<value_t> values = actsec.get_data();//and get all values
    vector<value_t>::iterator val = values.begin();//get first value
    for (vector<string>::iterator name = valuenames.begin(); name != valuenames.end(); name++, val++)//run trough value names and values simultaneously
      fprintf(f, "%s=%s\n", name -> c_str(), val -> stringvalue.c_str());//write this value :"valuname=value"
  }

  if (fclose(f)) throw new setupmanager_e(2, "file not closed: " + file);//it is not closed
}

const string &setupmanager::getstring(const string &section, const string &name) const {
  try {
    return sections[section][name].stringvalue;//find the section, in it the value and return the string one
  }
  catch (tree_e *e) {
    delete e;
    throw new setupmanager_e(10, "section or value not found");
  }
}

int setupmanager::getint(const string &section, const string &name) const {
  try {
    const value_t &val = sections[section][name];//get the right one
    if (!val.isint) throw new setupmanager_e(9, "value \"" + name + "\" in section \"" + section +"\" is not int");//it is not the right type
    return val.intvalue;//return the int
  }
  catch (tree_e *e) {
    delete e;
    throw new setupmanager_e(10, "section or value not found");
  }
}

float setupmanager::getfloat(const string &section, const string &name) const {
  try {
    const value_t &val = sections[section][name];//get the value
    if (!val.isfloat) throw new setupmanager_e(9, "value \"" + name + "\" in section \"" + section +"\" is not float");//it is not the right type
    return val.floatvalue;
  }
  catch (tree_e *e) {
    delete e;
    throw new setupmanager_e(10, "section or value not found");
  }
}

bool setupmanager::getbool(const string &section, const string &name) const {
  try {
    const value_t &val = sections[section][name];//get the value
    if (!val.isbool) throw new setupmanager_e(9, "value \"" + name + "\" in section \"" + section +"\" is not float");
    return val.boolvalue;//and return it
  }
  catch (tree_e *e) {
    delete e;
    throw new setupmanager_e(10, "section or value not found");
  }
}

void setupmanager::setstring(const string &section, const string &name, const string &value) {
  value_t *val = tryget(section, name);//get the value to change
  *val = processvalue(value);//and fill it with the value
}

void setupmanager::setint(const string &section, const string &name, int value) {
  value_t *val = tryget(section, name);//get the value to change
  val -> isint = true;//setting int, so it is int
  val -> intvalue = value;//set the int
  val -> isfloat = true;//int can be float as well
  val -> floatvalue = value;//so set it
  val -> isbool = false;//this is C, however, int is not taken as a bool
  char theint[24];//23 numerals should be enough for that int
  sprintf(theint, "%i", value);//convert to string
  val -> stringvalue = theint;//and save it there
}

void setupmanager::setfloat(const string &section, const string &name, float value) {
  value_t *val = tryget(section, name);//get the value to change
  val -> isint = false;//the float can't be converted to int
  val -> isbool = false;//float isn't bool
  val -> isfloat = true;//setting float
  val -> floatvalue = value;//set the value
  char thefloat[100];//it is long enough
  sprintf(thefloat, "%g", value);//convert, let it think what is best
  val -> stringvalue = thefloat;//save it
}

void setupmanager::setbool(const string &section, const string &name, bool value) {
  value_t *val = tryget(section, name);//get the value to change
  val -> isint = false;//bool is not float
  val -> isfloat = false;//bool is not a float as well
  val -> isbool = true;//nothing to add
  val -> boolvalue = value;//set the value
  if (value) val -> stringvalue = "true"; else val -> stringvalue = "false";//so, set the string as well
}

setupmanager::value_t *setupmanager::tryget(const string &sectionname, const string &name) {
  section *sec = sections.try_get(sectionname);//try to get the section
  value_t *val;//the value pointer
  if (!sec) {//the section doesn't exist, so create it and give it the value
    section newsec;//new empty section
    value_t newval;//I does not care, what is there, it will be changed soon
    newsec.add(name, newval);//the section won't be empty now
    val = &newsec[name];//get the pointer to the newval
    sections.add(sectionname, newsec);//add the new section
    //I won't get pointer to sec, it won't be needed now
  } else val = sec -> try_get(name);//if the section is there, find there the value to change
  if (!val) {//the value is not in the section yet
    value_t newval;//new value, doesn't care what's inside..
    sec -> add(name, newval);//add it there
                             //sec may not be NULL - if it existed before, nothing to care, and if it is new, it already have the val}
    val = sec -> try_get(name);//and get pointer to it
  }
  return val;//so, done
}

void setupmanager::clear() {
  sections.clear();//remove all sections with it's insides
}

void setupmanager::remove(const string &section_name, const string &name) {
  try {
    section &sec = sections[section_name];//get the section
    sec.remove(name);//remove the value
    if (sec.empty()) sections.remove(section_name);//the section is empty, there is no need of it
  }
  catch (tree_e *e) {
    delete e;
    throw new setupmanager_e(10, "section or value not found");
  }
}

setupmanager::value_t setupmanager::processvalue(const string &value) const {
  value_t res;//creating this value
  res.stringvalue = value;//copy the string value
  if ((value == "true") || (value == "yes") || (value == "Y")) {
    res.boolvalue = true;//it means true, so it is bool
    res.isbool = true;
    res.isint = false;//bool is not int nor float
    res.isfloat = false;
  } else if ((value == "false") || (value == "no") || (value == "N")) {
    res.boolvalue = false;
    res.isbool = true;//false is bool
    res.isint = false;//but it isn't float nor int
    res.isfloat = false;
  } else {
    res.isbool = false;
    char *ok;//here will be how it worked
    const char *val = value.c_str();//copy the string here
    res.floatvalue = strtod(val, &ok);//convert it
    res.isfloat = (ok != val);//did the conversion work?
    if (res.isfloat) {//if it is float it can be int as well
      res.intvalue = strtol(val, &ok, 10);//converting to int
      res.isint = (ok != val);//could it be converted?
    }
  }
  return (res);
}

stringlist setupmanager::getSectionList() const {
	return stringlist(sections.get_indexes());//get the list of indexes and convert it into stringlist
}

stringlist setupmanager::getEntryList(const std::string &section) const {
	try {
		return stringlist(sections[section].get_indexes());//get the right section, get its indexes and convert it
	}
	catch (tree_e *e) {//it is not there
		if (e->id == 1) {//OK, it is the known one
			delete e;//delete this exception
			throw new setupmanager_e(10, "section or value not found");//and replace it bo some other
		} else
			throw e;//it is unknown, just let it go
	}
}

}
