/***************************************************************************
 *   Copyright (C) 2005 by Michal Vaner                                    *
 *   vorner@centrum.cz                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef VORLIB_SOCKET_H
#define VORLIB_SOCKET_H

#include <string>//the std::string class
#include <sys/types.h>

#ifdef VORLIB_WINDOWS
#include <winsock2.h>
#endif

#include "exception.h"//it will throw exceptions, so it has to know them
#include "searchtree.h"//for indexing

namespace vorlib {

class socket_e : public exception {
public:
  socket_e(int id = 0, const std::string &message = "", int handle = -1);//this is default constructor
  //it can easilly use the default destructor
  int handle;//handle of socket that caused it (if it makes sence) or -1
};//the socket exception class

/*
ERROR IDs:
  -1: internal, errors in the class, some of them works only with DEBUG
  1: socket already connected, trying to connect it twice (TCP)
  2: could not allocate the handle
  3: trying to disconnect socket, that is not connected (TCP)
  4: couldn't convert host name to ip address (maybe it's wrongly spelled?)
  5: could not bind the socket
  6: socket not connected
  7: some data not sent yet, resend
  8: send error (something odd happened)
  9: no data to send
  10: receive error
  11: closed by peer, this end is closed automatically
  12: didn't read enough data, for some reason
  13: invalid operation with blocking/unblocking socket
  14: error while accepting connection
  15: would block while trying to accept, no connection waiting on non-blocking socket
  16: no remote end with listening port
  17: trying to copy server, YOU CAN NOT DO THAT
  18: error while connecting
  19: select error
  20: console is already scanned
  21: error reading from console
  22: handle is not my socket
  23: imposible request
  24: reset by peer - this does some windows applications before actually closing the socket, usually means nothing and should be ignored
  25: error reading socket options/status
  26: connect error, FATAL, socket removed
  27: not a udp socket
  28: unknown address family
  29: source address not ready, nothing came yet
*/

class base_socket {
private:
protected:
  bool hassock;//do we have a socket allocated
  bool connected;//are we connected?
  int sockethandle;//the identificator of socket
public:
  base_socket();//constructor, creates unconected socket
  virtual ~base_socket();//destructor, removes the socket
  virtual void write(const void *data, int size) = 0;//sends some raw data
  virtual int read(void *data, int max) = 0;//it reads into data eather how much is availible or to max bytes and returns read. data must be allocated already
  virtual int readblocking(void *data, int max) = 0;//read data and wait for them, if there are no data availible
  virtual void readexact(void *data, int size) = 0;//reads size of data, waits for all them, always returns this much of them
  bool isconnected();//ask, weather the socket is connected
  int getsocket();//returns the socket handle, for select or another use. PLEASE, do not connect, close or something like this with it
};
/*
this is the base socket class
it is ABSTRACT, it can't be used directly
it is recomendet to use text comunication or use conversion functions (like htons) becouse of the big and little endian
*/

class tcp_socket : public base_socket {
private:
  void preopen();//preparing function to connection
  void open(void *addr);//well, it should be sockaddr, but there are different headers under win and lin, so I wont include them here
  void *datawaiting;//what data are waiting
  int datalen;//length of data
  int datasent;//what data are sent
  bool block;//is it allowed to block?
  bool listening;//are we listening
protected:
public:
  tcp_socket();//constructor, UNCONNECTED
  tcp_socket(bool block);//is the socket allowed to block?
  virtual ~tcp_socket();//destructor
  void open(unsigned addr, int port);//connect by ip adrress and port (in native host format, it will be converted to network in this func)
  void open(const std::string &addr, int port);//connect by name and port (in native host format)
  void disconnect();//close the connection and wait for another
  void listenonport(int port);//listen on that port
  void listenonport(int port, unsigned allowaddr, unsigned char maxqueue);//some more setup than above
  tcp_socket *acceptconnection();//accepts incoming connection, returning new socket (already connected), ALLOCATES THE SOCKET
  tcp_socket *acceptconnection(unsigned &addr, int &port);//get the port and address as well
  virtual void write(const void *data, int size);//send some data, if ends with socket_e/7, only some of them were sent and you need to resend them (func resend), or close socket
  void resend();//send the rest, if it couldn't be sent in one try
  void resetwaiting();//removes all waiting data
  bool iswaiting();//ask, weather some data waits
  void addwaiting(const void *data, int size);//add some data for waiting, then use resend
  virtual int read(void *data, int max);//get some data (U have to allocate that before), non-blocking
  virtual int readblocking(void *data, int max);//same as above, if no data availible, waits for them (and may wait a LONG time), can not with non blocking socket
  virtual void readexact(void *data, int size);//read exactly size of data, can cause socket_e 12, but the data will still be there, can not with non blocking socket
};
/*
this is socket working with TCP/IP protocol
it has to be connected before using
it can be connected more times (but has to be disconnected before)
The socket can not be coppyed, use references instead
*/

typedef enum {
#ifndef VORLIB_NO_IPV6
  af_ipv6,//does we support the ipv6?
#endif
  af_ipv4
} address_family;//witch addresse do it use?

#define ADDR_LEN 27

typedef struct {
  unsigned short af;//the address family
  char data[ADDR_LEN];//the 2 bytes for protocol and 16 for the address and there are two other 4byte vars, witch I don't know what are they for
} long_sockaddr;//the ipv6 version would not fit into the standart one

typedef struct {
  unsigned short af;//the address family, here AF_INET
  unsigned short port;//the port number
  unsigned address;//four bytes the address
  char zero[ADDR_LEN - 6];//the rest
} ipv4_addr;

#ifndef VORLIB_NO_IPV6

typedef struct {
  unsigned short af;//the address family
  unsigned short port;//the port
  unsigned flow_info;//something odd, I don't know what's that for, however, it shall be there
  unsigned char address[16];//the address is 16 bytes large
  unsigned scope_info;//I don't know whats that for
  char zero[ADDR_LEN - 26];//to make sure, it will be large enough, if I ever make the address bigger
} ipv6_addr;//the address

#endif

class address {
private:
  long_sockaddr addr;//the own address
public:
  address();//default constructor, however, it should not be used, because noone will ever know, what will be inside
  address(const std::string &address, int port);//create it with address as string, weather it will be ipv4 or 6 doesn't matter (it will be 4, if it can and it can be accepted by any computer)
  address(char _1, char _2, char _3, char _4, int port);//get the address (ipv4) as the four point separated bytes
  address(unsigned addr, int port);//the ipv4 address as one unsigned
  address(const long_sockaddr *addr, int length = ADDR_LEN + 2);//get the address directly from this, however, only the supported types can be given
  address_family get_af() const;//get the address family
  int get_port() const;//returns the port
  const long_sockaddr *get_addr() const;//get the raw address
  std::string get_str_addr() const;//get the address turned into a string (not the host name, just the dot formet)
  unsigned get_un_addr() const;//return it as unsigned, however, it works only for ipv4
};//this class manages addresses of computers, at this moment ipv4 and ipv6 addresses
//all addresses are in host format, if it matters, except for the sockaddr's, they are in network byte order

address getaddr(const std::string &addr, int port);//it will be decided automatically, weather it will be ipv4 or ipv6
address getaddr(unsigned addr, int port);
address getaddr(char _1, char _2, char _3, char _4, int port);
address getaddr(const long_sockaddr *addr, int length = ADDR_LEN + 2);
//use theese functions, if you want to pass it as an argument to some function without making your own wariable, so it will be freed automatically
//if you want to listen on a localhost with ipv6 (to allow both 4 and 6 versions to come to you), use "::1", witch means localhost in ipv6

typedef enum {
  sst_none,//no socket at all
  sst_tcp_listen,//for the tcp listening port
  sst_tcp_waiting_connect,//this is tcp socket that is still waiting for connection
  sst_tcp_transfer,//data transfer socket
  sst_line_transfer,//line transfer socket
  sst_console,//char by char console
  sst_line_console,//line console
  sst_udp//udp socket
} server_socket_type;

typedef enum {
  dsr_local_closed, //locally closed socket
  dsr_remote_closed, //the rome host closed it
  dsr_error //closed becouse of error (or not created...) - timeout or something like this
} delsocket_reason;

class server;//say, it will be there somewhere later

class server_socket {
private:
public:
  server_socket();//constructor
  virtual ~server_socket();//destructor
  virtual server_socket_type gettype() const = 0;//returns type of socket
  virtual address getremoteaddr() const = 0;//return remote address
  virtual int gethandle() const = 0;//returns handle of the socket
  virtual bool wantwrite() const= 0;//this one tells, weather the socket has something to be written
  virtual bool canread() const = 0;//Is there a posibility I could read from U, if there would be something?
  virtual void write(int size, const char *data) = 0;//puts some data into queue
  virtual bool socket_ready() = 0;//tell the socket to write as many data as it can, return true, when the queue is empty
  virtual bool queue_empty() const = 0;//returns weather the queue is empty or not
  virtual int queue_size() const = 0;//return the size of that queue
  bool to_delete;//is it about to be deleted? - it is not deleted imediatelly
};//this is one socket of this server below, virtual one. It is no goot alone, used by server, DO NOT USE IT

//the server parameter contains the server, that caused the event
typedef void (*errorevent) (const socket_e *, server *);//this is header of function that is called, when an error eccures
//at the time, errno should still contain the error
typedef void (*timerevent) (server *);//this is called when a timer has expired
typedef void (*consolereadyevent) (server *);//this will be called when there is something on console, read it yourself, will be called until you read it
typedef void (*consolelineevent) (const std::string &, server *);//this is called with a line from console
typedef server_socket_type (*tcplistenevent) (const address &addr, int orig_sock, int new_sock, server *);//this will be called when new connection arrives, addr and port specify the remote address of the caller, orig_sock is handle of the listening socket and the new_sock is the new one just accepted
//return sst_tcp_transfer fo standart (data) tcp socket
//or sst_line_transfer fo line-handled tcp soket
//or sst_none if you don't want this connection. Then, it will be closed and the remote host will get message 'REJECTED' trough this socket
typedef void (*newsocketevent) (int socket, server_socket_type type, server *);//this eevent happens every time a socket is created
typedef void (*delsocketevent) (int socket, server_socket_type type, delsocket_reason reason, server *);//called every time a socket is closed/removed
typedef void (*dataarriveevent) (int socket, const char *buffer, int size, server *);//this will be called when some data arrive trough socket (not console)
//socket is the handle of socket that brough it
//if it is transfer socket, buffer contains size of bytes of data (and can be, of course, retyped to void *)
//if it is line socket, it will contain one line (without the \n and with the NUL char at the end)  and size will be length of the line without the NUL character
typedef void (*queueemptyevent) (int socket, server *);//this is called when some socket's queue is empty
typedef void (*datasentevent) (int socket, int queue_size, server *);//this is called when some data from some socket are sent out, providing the resting size of queue
//the difference between these two is, that queueempty is called only, when the queue becomes empty adn the datasentevent every time some data from queue are sent

class server {
private:
  int socketcount;//count of socket used
  int prealloc;//how much are allocated?
  server_socket **sockets;//array of all sockets
  int addsock(server_socket *sock);//add this to list of sockets
  void delsock(int socket, delsocket_reason = dsr_local_closed);//remove this socket from list
  bool running;//is the server terminated, shall it stop going?
  void rebuild_lists();//rebuild lists of the server
  fd_set readsocks;//the read sockets
  fd_set writesocks;//the write sockets
  void checkall() const;//check everything is OK in the tree and so on
  int maxfd;
  typedef struct {
    server_socket *socket;//the own socket
    int index;//the index of socket
  } socket_ident;//ident of the socket
  balanced_search_tree<int, socket_ident> indexer;//index of all sockets, by the handle, for faster lookup
  bool checkevents(int count);//happened anything?
  void raiseexcp(socket_e *);//just tell the user, that there was some exception
  void delete_all();//delete all marked sockets
  void test_for_udp(int index);//test for udp and if it is not a udp socket, it throws an exception

  errorevent error_event;//this is pointer to the error event function
  newsocketevent new_socket_event;//called when new socket is created
  delsocketevent del_socket_event;//called when socket dies

  int timer_interval;//the interval of timer, in seconds
  time_t time_next;//when it will happen again?
  timerevent timer_event;//the timer event
  void reset_timer();//generates new setup of timer

  consolereadyevent console_ready_event;
  consolelineevent console_line_event;
  dataarriveevent data_arrive_event;
  queueemptyevent queue_empty_event;//when the queue becomes empty
  datasentevent data_sent_event;//when some data are sent out

  tcplistenevent tcp_listen_event;
public:
  server();//constructor
  server(const server &);//this is copy constructor, jut for throwing and exception. YOU CAN NOT COPY THIS
  virtual ~server();//destructor
  void operator=(const server &);//this is assignment operator. Throws exception, YOU CAN NOT COPY THIS
  int tcp_listen(int port, address_family af = af_ipv4);//creates tcp listening socket
  int tcp_listen(const address &addr, unsigned char maxqueue = 10);//set the address to port and address you want to allw or INETADDR_ANY, if anyone can connect
  //this one does not mean it is connected, event when it did not throw error, it may be waiting to the remote ent to accept
  //you can write to it's queue, however nothing will be sent until it is connectde
  //the best way to know it connected already is that it seys it's queue is empty(if you did not write to it)
  int tcp_connect(const address &addr);//connect to the "addr" endpoint, crating a tcp_transfer_socket
  int line_connect(const address &addr);//this is tcp socket that is used for line protocols
  int add_udp_socket(address_family af = af_ipv4);//creates new udp socket
  void set_udp_local(int sock, int port);//sets local udp port, it is set on ipv6 by default, if it is enabled
  void set_udp_destination(int sock, const address &addr);//set the address, where the packets will be sent
  address get_udp_source(int sock);//return address of the last incoming packet
  //all theese returns id of the socket
  int add_console();//check, weather the console input is ready
  int add_line_console();//will cause lineready event instead dataready, the difference is that this wait until there is whole line
  //you can have only one console checker
  //theese two WONT work under windows
  void close_socket(int socket);//remove the socket from there

  void terminate();//tell the server to terminate
  void go(int refresh = 1);//start the server, refresh rate in secs (if something comes trough socket, it reacs imediatelly
  //it is said, that on Windows, the select does not work properly with no sockets, so it may not work on Windows if the server is entirely "empty"

  void set_onerror(errorevent);//set the error event
  errorevent get_onerror() const;//get the set value
  //if you want the server stop because of the exception, just call terminate()
  //this is called only when the server is running (inside the go() function), not in startup or something like that
  //if this is not set (or set to NULL), the server will terminate imediatelly, leaving exception behind
  //do not throw exceptions here, it may cause a resource leak
  void set_onnewsocket(newsocketevent);//set the function that is called when new socket comes
  newsocketevent get_onnewsocket() const;
  void set_ondelsocket(delsocketevent);//this one when a socket dies
  delsocketevent get_ondelsocket() const;//on delsock is called before it is actually removed, so it is posible to access it in this event (to found out type or so), but may not be writeable any more

  void set_ontimer(timerevent);//set the timer event
  timerevent get_ontimer() const;//get the timer value
  void set_interval(int);//set the interval of timer
  int get_interval() const;//get the interval
  //you can have one timer in the server, with setupable interval
  //if you set the timeevent to NULL, it will never come
  //the interval is in seconds, default 30
  void set_onconsoleready(consolereadyevent);//set the function that will be called, when console has some data. May cause problems with buffered console, better to use line console
  consolereadyevent get_onconsoleready() const;//say, witch function is set now
  //to enable scanning console, you have to add_console();
  //however, you can have only one console
  void set_onconsoleline(consolelineevent);
  consolelineevent get_onconsoleline() const;//theese two are for line console
  void set_ondataarrive(dataarriveevent);//witch func will be called when some data come?
  dataarriveevent get_ondataarrive() const;

  void set_ontcplisten(tcplistenevent);//set the function for accept new connections
  tcplistenevent get_ontcplisten() const;//return tha actually set function

  server_socket_type get_type(int socket) const;//returns type of that socket
  address get_remote_addr(int socket) const;//returns remote address of that socket (in the case of the udp socket, it is the destination)
  int get_sock_count() const;//returns count of sockets
  int get_socket(int index) const;//returns handle of socket (index 0..get_sock_count - 1). This can change!!, only for getting list of all sockets
  int operator[](int index) const;//same as above
  bool queue_empty(int socket) const;//is the write queue empty on that socket?
  int queue_size(int socket) const;//ask, how many bytes are inside the queue
  void set_onqueueempty(queueemptyevent);//this one will be called when some socket's write queue becomes empty (all data are sent)
  queueemptyevent get_onqueueempty() const;//ask, what function will be called...
  void set_ondatasent(datasentevent);//this is called when some data were sent out
  datasentevent get_ondatasent() const;

  void write_data(int socket, int size, const void *data);//write these data into socket
  void write_data(int socket, const char *format, ...);//write this string, the NULL at the end will be omited, works up to 10K string (after processing)

  void *add_data;//some more place for any data you want to bind with the server, server does not care a bit about this
};
/*
  This class can manage many tcp and udp sockets, console input... It works upon text protocols, sending and receiving lines, or data protocols
  Create it, setup some basic sockets and call run. It will run until you call terminate and will call your functions
  It can manage whole server or can be forked and used as easy to use client. However, this one does NOT use multithreading
  DO NOT COPY IT (it won't let you anyway), just use references
  I hope it to make it manage pipes and files one time.
*/

}

#endif
