/***************************************************************************
 *   Copyright (C) 2005 by Michal Vaner                                    *
 *   vorner@centrum.cz                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stringlist.h"

namespace vorlib {

stringlist::stringlist() {
  string_count = 0;
  strings = NULL;//we have none
}

stringlist::~stringlist() {
  string **end_str = strings + string_count;
  for (string **str = strings; str != end_str; str++) delete (*str);//free all strings
  free (strings);//and free the array
}

stringlist::stringlist(const stringlist& source) {
  string_count = source.string_count;//how many of strings do we take?
  strings = (string **) malloc (string_count * sizeof *strings);//allocate the array
  string **end_str = strings + string_count;//one after last
  string **src = source.strings;//begin of source array
  for (string **str = strings; str != end_str; str++, src++) *str = new string (**src);//copy them all
}

stringlist::stringlist(const std::vector<std::string> &source) {
	string_count = source.size();
	strings = (string **) malloc(string_count *sizeof *strings);
	string **end_str = strings + string_count;
	std::vector<std::string>::const_iterator it = source.begin();
	for (string **str = strings; str != end_str; it++, str++)
		*str = new string(*it);
}

stringlist &stringlist::operator=(stringlist &source) {
  string **end_str = strings + string_count;
  for (string **str = strings; str != end_str; str++) delete (*str);//delete all actual strings
  string_count = source.string_count;
  strings = (string **) realloc (strings, string_count * sizeof *strings);//new size of the array
  end_str = strings + string_count;//the end probably moved
  string **src = source.strings;
  for (string **str = strings; str != end_str; src++, src++) *str = new string (**src);//copy the strings
  return (*this);
}

int stringlist::size() const{
  return (string_count);
}

string &stringlist::operator[](int index) {
  return (*strings[index]);
}

const string &stringlist::operator[](int index) const{
  return (*strings[index]);
}

void stringlist::add(const string &value) {
  strings = (string **) realloc (strings, ++string_count * sizeof *strings);//increase the size of array
  strings[string_count - 1] = new string (value);//copy there the string
}

stringlist &stringlist::operator+=(const string &value) {
  add(value);//just add it to me :)
  return (*this);//and return the new value
}

void stringlist::join(const stringlist &source) {
  int act_count = string_count;
  strings = (string **) realloc (strings, (string_count += source.string_count) * sizeof *strings);
   //increase the size of array by count of new strings
  string **end_str = strings + string_count, **src = source.strings;//last + 1 of my array, first of source array
  for (string **str = strings + act_count; str != end_str; str++, src++) *str = new string (**src);//copy the new strings
}

stringlist &stringlist::operator+=(const stringlist &source) {
  join(source);//copy them to me
  return (*this);//and return the new value
}

stringlist &stringlist::operator+(const string &value) const {
  stringlist *result = new stringlist (*this);
  *result += value;
  return (*result);
}

stringlist &stringlist::operator+(const stringlist &source) const {
  stringlist *result = new stringlist (*this);
  *result += source;
  return (*result);
}

void stringlist::insert(int index, const string &value) {
  if (index < 0) index = 0;
  if (index > string_count) index = string_count;//put it into range
  strings = (string **) realloc (strings, ++string_count * sizeof *strings);//get one more place
  string **stop = strings + index;
  for (string **str = strings + string_count - 1; str != stop; str--) *str = *(str - 1);//move them one forwart to the end
  strings[index] = new string (value);//copy there the new one
}

bool stringlist::remove(const string &value) {
  return (remove(find(value)));//find the item, remove it and return weather it was there
}

bool stringlist::remove(int index) {
  if ((index < 0) || (index >= string_count)) return (false);//it is not there
  delete (strings[index]);//free the one we are talking about
  string **end = strings + string_count - 1;//to the last
  for (string **str = strings + index; str != end; str++) *str = *(str + 1);//move them into the hole
  strings = (string **) realloc (strings, --string_count * sizeof *strings);//shrink the array by one
  return (true);
}

bool stringlist::remove_last() {
  if (string_count == 0) return (false);
  delete (strings[string_count - 1]);//free the last one
  strings = (string **) realloc (strings, --string_count * sizeof *strings);//shring the last one off
  return (true);
}

void stringlist::clear() {
  string **end_str = strings + string_count;
  for (string **str = strings; str != end_str; str++) delete (*str);//free all strings
  free (strings);//free the array
  string_count = 0;
  strings = NULL;//and say it's empty
}

int stringlist::count() const {
  return (string_count);
}

int stringlist::find(const string &value) const {
  string **end_str = strings + string_count;
  for (string **str = strings; str != end_str; str++) if ((**str) == value) return (str - strings);//is it the one?
  return (-1);//nothing found
}

int stringlist::operator%(const string &value) const {
  return (find(value));
}

stringlist &stringlist::operator-=(int index) {
  remove(index);//remove it
  return (*this);//and return myself
}

stringlist &stringlist::operator-=(const string &value) {
  remove(value);
  return (*this);
}

stringlist &stringlist::operator-(int index) const{
  stringlist *result = new stringlist (*this);//copy me
  (*result) -= index;//remove the thing
  return (*result);//and return it
}

stringlist &stringlist::operator-(const string &value) const {
  stringlist *result = new stringlist (*this);
  (*result) -= value;
  return (*result);
}

}
